﻿/// @copyright  	Copyright (c) 2014 Toradex AG.
///                 [Software License Agreement]
/// $Author:		khv $
/// $Date: 			$ 2014-06-18 $
/// @brief:			Registry access demo in VC#

using System;
using System.Linq;
using System.Collections.Generic;
using System.ComponentModel;
using System.Data;
using System.Drawing;
using System.Text;
using System.Windows.Forms;
using System.Security.Cryptography;
using System.Runtime.InteropServices;

namespace registry_access_demo
{
    public partial class Form1 : Form
    {
        [DllImport("coredll.dll", CharSet = CharSet.Auto, SetLastError = true)]
        public static extern uint RegOpenKeyEx(
           UIntPtr hKey,
           string lpSubKey, 
           uint ulOptions,
           uint samDesired, 
           out UIntPtr phkResult);

        [DllImport("coredll.dll", CharSet = CharSet.Auto, SetLastError = true)]
        public static extern uint RegCreateKeyEx(
           UIntPtr hKey,
           string lpSubKey,
           uint dwReserved,
           string lpClass,
           uint dwOptions, 
           uint samDesired, 
           UIntPtr lpSecurityAttributes,
           out UIntPtr phkResult, 
           out uint lpdwDisposition);

        [DllImport("coredll.dll", CharSet = CharSet.Auto, SetLastError = true)]
        public static extern uint RegQueryValueEx(
            UIntPtr hkey,
            String lpValueName,
            UIntPtr lpReserved,
            ref uint lpType,
            byte[] lpData,
            ref uint lpcbData);

        [DllImport("coredll.dll", CharSet = CharSet.Auto, SetLastError = true)]
        public static extern uint RegSetValueEx(
            UIntPtr hKey,
            String lpValueName,
            uint dwReserved,
            uint dwType,
            byte[] lpData,
            uint cbData);

        [DllImport("coredll.dll", CharSet = CharSet.Auto, SetLastError = true)]
        public static extern uint RegDeleteKey(UIntPtr hKey, string lpSubKey);

        [DllImport("coredll.dll", CharSet = CharSet.Auto, SetLastError = true)]
        public static extern uint RegDeleteValue(UIntPtr hKey, string lpValueName);

        [DllImport("coredll.dll", CharSet = CharSet.Auto, SetLastError = true)]
        public static extern uint RegFlushKey(UIntPtr hKey);

        [DllImport("coredll.dll", CharSet = CharSet.Auto, SetLastError = true)]
        public static extern uint RegCloseKey(UIntPtr hKey);

        public const uint MAX_BUFFER_SIZE = (uint)256;

        public UIntPtr HKEY_CLASSES_ROOT    = (UIntPtr)0x80000000;
        public UIntPtr HKEY_CURRENT_USER    = (UIntPtr)0x80000001;
        public UIntPtr HKEY_LOCAL_MACHINE   = (UIntPtr)0x80000002;
        public UIntPtr HKEY_USERS           = (UIntPtr)0x80000003;

        public const uint KEY_READ                  = (uint)0x00020019;
        public const uint KEY_WRITE                 = (uint)0x00020006;
        public const uint KEY_ALL_ACCESS            = (uint)0x00FF003F;
        public const uint ERROR_SUCCESS             = (uint)0;
        public const uint REG_OPTION_VOLATILE       = (uint)1;
        public const uint REG_OPTION_NON_VOLATILE   = (uint)0;

        public const uint REG_SZ          = (uint)1;
        public const uint REG_BINARY      = (uint)3;
        public const uint REG_DWORD       = (uint)4;
        public const uint REG_MULTI_SZ    = (uint)7;

        string operationToPerform = "";
        UIntPtr mainKeySelect = (UIntPtr)0x00000000;
        UIntPtr retOpenKey = UIntPtr.Zero;
        UIntPtr retNewKey = UIntPtr.Zero;
        uint retVal = 0;
        string subKeyPath = "";
        string keyValueName = "";
        uint disposition = 0;
        uint valueType = (uint)0;
        byte[] writeRegData = new byte[MAX_BUFFER_SIZE];
        uint dataLength = (uint)0;
        bool allKeysClosed = false;
        byte[] readRegData = new byte[MAX_BUFFER_SIZE];

        /// Shows error code occurred in any procedure.
        public void ShowErrorCode(uint errorCode)
        {
            switch (errorCode)
            {
                case (uint)2:
                    {
                        if (operationToPerform == "Read a value" || operationToPerform == "Delete a value")
                            MessageBox.Show("Value not found.", "Error");
                        else
                            MessageBox.Show("Key not found.", "Error");
                        break;
                    }
                case (uint)5:
                    {
                        MessageBox.Show("Access denied.", "Error");
                        break;
                    }
                case (uint)87:
                    {
                        MessageBox.Show("Invalid parameter.", "Error");
                        break;
                    }
                case (uint)234:
                    {
                        MessageBox.Show("Data too large to handle.", "Error");
                        break;
                    }

                default:
                    {
                        MessageBox.Show("Error Code (winerror.h) = " + errorCode, "Error");
                        break;
                    }
            }
        }

        /// Converts an unsigned integer 32 bit value to byte array.
        public byte[] ConvertIntToBytes(uint intValue)
        {
            byte[] byteArray = new byte[4];                     ///< 32 bits = 4 bytes.
            int count = 3;

            while (count >= 0)
            {
                byteArray[count] = (byte)(intValue % 256);
                intValue >>= 8;
                count--;
            }
            return byteArray;
        }

        /// Takes data in string format (as in textbox) and converts that to byte array i.e. in registry writeable form.
        public byte[] GetBytesAndLength(string sourceString, uint valueType)
        {
            byte[] returnBytes = new byte[sourceString.Length * sizeof(char)];
            uint data = (uint)0;
                  
            switch(valueType)
            {
                case REG_SZ:
                case REG_MULTI_SZ:
                    {
                        System.Buffer.BlockCopy(sourceString.ToCharArray(), 0, returnBytes, 0, (sourceString.Length * 2));      ///< String to byte array conversion.
                        dataLength = (uint)((sourceString.Length * 2));      ///< Wide character length = Normal character length * 2.
                        break;
                    }
                case REG_DWORD:
                    {
                        data = Convert.ToUInt32(sourceString);              ///< String to DWORD conversion.
                        dataLength = (uint)4;                               ///< Unsigned int = 4 bytes.
                        returnBytes = BitConverter.GetBytes(data);          ///< Unsigned int to byte array conversion.
                        break;
                    }
                case REG_BINARY:
                    {
                        data = Convert.ToUInt32(sourceString);              ///< String to unsigned int conversion.
                        dataLength = 4;                                     ///< Unsigned int = 4 bytes.
                        returnBytes = ConvertIntToBytes(data);              ///< Separate 4 bytes of unsigned int.
                        break;
                    }
            }
            return returnBytes;
        }

        /// Shows data received by "Read a value" operation to the user.
        public void ShowReadData(byte[] databytes, uint dataType, uint dataLength)
        {
            uint byteArrayToInt = (uint)0;
            uint totalBytesToPrint = 0;
            uint dataCopied = 0;
            uint count = 2;
                        
            switch (dataType)
            {
                case REG_SZ:
                    {
                        ValueType.SelectedItem = "String";
                        Value.Text = Encoding.Unicode.GetString(databytes, 0, (int)dataLength);     ///< Byte array to string conversion.
                        break;
                    }
                case REG_DWORD:
                    {
                        ValueType.SelectedItem = "Dword";
                        byteArrayToInt = BitConverter.ToUInt32(databytes, 0);           ///< Byte array to unsigned int conversion.
                        Value.Text = byteArrayToInt.ToString();                         ///< Unsigned int to string conversion.
                        break;
                    }
                case REG_BINARY:
                    {
                        ValueType.SelectedItem = "Binary";
                        Array.Reverse(databytes, 0, 4);                                 ///< Reverse all 4 bytes in byte array.
                        /// byte[] reversed = bytes.Reverse().ToArray();
                        byteArrayToInt = BitConverter.ToUInt32(databytes, 0);           ///< Byte array to unsigned int conversion.
                        Value.Text = byteArrayToInt.ToString();                         ///< Unsigned int to string conversion.
                        break;
                    }
                case REG_MULTI_SZ:
                    {
                        ValueType.SelectedItem = "Multi String";
                        Value.Enabled = true;
                        Value.Text = Encoding.Unicode.GetString(databytes, 0, (int)dataLength);     ///< Byte array to string conversion.
                        while (totalBytesToPrint < dataLength)
                        {
                            Value.Text += Environment.NewLine;
                            totalBytesToPrint = (uint)(Value.Text.Length * 2);
                            dataCopied = totalBytesToPrint - count;
                            Value.Text += System.Text.Encoding.Unicode.GetString(databytes, (int)dataCopied, (int)(dataLength - dataCopied));
                            count += 2;
                        }
                        break;
                    }
            }
        }
    
        /// Verifies the presence of data in required fields to perform the selected operation.
        public bool VerifyData()
        {
            if (operationToPerform == "")
            {
                MessageBox.Show("Select operation to perform.", "Error");
                return false;
            }

            if (mainKeySelect == (UIntPtr)0)
            {
                MessageBox.Show("Select main key.", "Error");
                return false;
            }

            switch (operationToPerform)
            {
                case "Create new key" :
                case "Delete a key"   :
                case "Read a value"   :
                case "Delete a value" :
                    {
                        if (NewKeyValue.Text == "")
                        {
                            MessageBox.Show("Enter name of the key/value.", "Error");
                            return false;
                        }
                        break;
                    }
                case "Create new value":
                case "Modify a value":
                    {
                        if (NewKeyValue.Text == "")  
                        {
                            MessageBox.Show("Enter name of the value.", "Error");
                            return false;
                        }
                        if (valueType == (uint)0)
                        {
                            MessageBox.Show("Enter value type", "Error");
                            return false;
                        }
                        if (Value.Text.Length > 127)
                        {
                            MessageBox.Show("Value too large.", "Error");
                            return false;
                        }
                        if (ValueType.Text == "Dword" || ValueType.Text == "Binary")
                        {
                            try
                            {
                                Convert.ToUInt32(Value.Text);
                            }
                            catch
                            {
                                MessageBox.Show("Enter an integer value.", "Error");
                                return false;
                            }
                        }
                        break;
                    }
            }
            return true;
        }

        public Form1()
        {
            InitializeComponent();
        }

        private void ExecuteCommand_Click(object sender, EventArgs e)
        {
            bool verify = VerifyData();
            if (verify == false)
            {
                return;
            }
            subKeyPath = SubKey.Text;                                   ///< Gets sub key name from sub key textbox.
            keyValueName = NewKeyValue.Text;                            ///< Gets key/value name from key/value textbox.
                       
            switch (operationToPerform)
            {
                case "Create new key":
                    {
                        retVal = RegOpenKeyEx(mainKeySelect, subKeyPath, 0, KEY_ALL_ACCESS, out retOpenKey);        ///< Opens the key.
                        if (retVal != ERROR_SUCCESS)
                        {
                            ShowErrorCode(retVal);
                        }
                        else
                        {
                            allKeysClosed = false;
                            retVal = ERROR_SUCCESS;
                            retVal = RegCreateKeyEx(retOpenKey, keyValueName, (uint)0, "", REG_OPTION_NON_VOLATILE, (uint)0, UIntPtr.Zero, out retNewKey, out disposition);      ///< Creates new key.
                            if (retVal != ERROR_SUCCESS)
                            {
                                ShowErrorCode(retVal);
                            }
                            else
                            {
                                MessageBox.Show("  Key created.", "Success");
                            }
                        }
                        RegCloseKey(retNewKey);           ///< Closes the open key.
                        RegCloseKey(retOpenKey);          ///< Closes the open key.
                        allKeysClosed = true;
                        break;
                    }
                case "Delete a key":
                    {
                        retVal = RegOpenKeyEx(mainKeySelect, subKeyPath, 0, KEY_ALL_ACCESS, out retOpenKey);        ///< Opens the key.
                        if (retVal != ERROR_SUCCESS)
                        {
                            ShowErrorCode(retVal);
                        }
                        else
                        {
                            allKeysClosed = false;
                            retVal = ERROR_SUCCESS;
                            retVal = RegDeleteKey(retOpenKey, keyValueName);            ///< Deletes the key.
                            if (retVal != ERROR_SUCCESS)
                            {
                                ShowErrorCode(retVal);
                            }
                            else
                            {
                                MessageBox.Show("  Key deleted.", "Success");
                            }
                        }
                        RegCloseKey(retOpenKey);            ///< Closes the open key.
                        allKeysClosed = true;
                        break;
                    }
                case "Create new value":
                case "Modify a value":
                    {
                        retVal = RegOpenKeyEx(mainKeySelect, subKeyPath, 0, KEY_ALL_ACCESS, out retOpenKey);        ///< Opens the key.
                        if (retVal != ERROR_SUCCESS)
                        {
                            ShowErrorCode(retVal);
                        }
                        else
                        {
                            allKeysClosed = false;
                            retVal = ERROR_SUCCESS;
                            writeRegData = GetBytesAndLength(Value.Text, valueType);    ///< Converts data given by user to registry writeable from.
                            retVal = RegSetValueEx(retOpenKey, keyValueName, (uint)0, valueType, writeRegData, dataLength);        ///< Creates/Modifies the value.
                            if (retVal != ERROR_SUCCESS)
                            {
                                ShowErrorCode(retVal);
                            }
                            else
                            {
                                MessageBox.Show("Modification done.", "Success");
                            }
                        }
                        RegCloseKey(retOpenKey);            ///< Closes the open key.
                        allKeysClosed = true;
                        break;
                    }
                case "Read a value":
                    {
                        retVal = RegOpenKeyEx(mainKeySelect, subKeyPath, 0, KEY_ALL_ACCESS, out retOpenKey);        ///< Opens the key.
                        if (retVal != ERROR_SUCCESS)
                        {
                            ShowErrorCode(retVal);
                        }
                        else
                        {
                            allKeysClosed = false;
                            retVal = ERROR_SUCCESS;
                            dataLength = MAX_BUFFER_SIZE;                         ///< Size of buffer to pass to RegQueryValueEx()
                            retVal = RegQueryValueEx(retOpenKey, keyValueName, UIntPtr.Zero, ref valueType, readRegData, ref dataLength);           ///< Reads the value.
                            if (retVal != ERROR_SUCCESS)
                            {
                                ShowErrorCode(retVal);
                            }
                            else
                            {
                                ShowReadData(readRegData, valueType, dataLength);       ///< Shows read value from registry to the user.
                            }
                        }
                        RegCloseKey(retOpenKey);        ///< Closes the open key.
                        allKeysClosed = true;
                        break;
                    }
                case "Delete a value":
                    {
                        retVal = RegOpenKeyEx(mainKeySelect, subKeyPath, 0, KEY_ALL_ACCESS, out retOpenKey);        ///< Opens the key.
                        if (retVal != ERROR_SUCCESS)
                        {
                            ShowErrorCode(retVal);
                        }
                        else
                        {
                            allKeysClosed = false;
                            retVal = ERROR_SUCCESS;
                            retVal = RegDeleteValue(retOpenKey, keyValueName);          ///< Deletes the value.
                            if (retVal != ERROR_SUCCESS)
                            {
                                ShowErrorCode(retVal);
                            }
                            else
                            {
                                MessageBox.Show("Value deleted.", "Success");
                            }
                        }
                        RegCloseKey(retOpenKey);            ///< Closes the open key.
                        allKeysClosed = true;
                        break;
                    }
            }
        }

        private void MainSelection_SelectedIndexChanged(object sender, EventArgs e)
        {
            if (MainSelection.Text == "Create new key")
            {
                operationToPerform = "Create new key";
                ValueType.Enabled = false;
                Value.Enabled = false;
            }
            else if (MainSelection.Text == "Delete a key")
            {
                operationToPerform = "Delete a key";
                ValueType.Enabled = false;
                Value.Enabled = false;
            }
            else if (MainSelection.Text == "Create new value")
            {
                operationToPerform = "Create new value";
                ValueType.Enabled = true;
                Value.Enabled = true;
            }
            else if (MainSelection.Text == "Read a value")
            {
                operationToPerform = "Read a value";
                ValueType.Enabled = false;
                Value.Enabled = false;
                ValueLabel.Text = "Value";
                Value.Text = "";
            }
            else if (MainSelection.Text == "Modify a value")
            {
                operationToPerform = "Modify a value";
                ValueType.Enabled = true;
                Value.Enabled = true;
            }
            else if (MainSelection.Text == "Delete a value")
            {
                operationToPerform = "Delete a value";
                ValueType.Enabled = false;
                Value.Enabled = false;
            }
        }
        
        private void MainKey_SelectedIndexChanged(object sender, EventArgs e)
        {
            if (MainKey.Text == "HKEY_CLASSES_ROOT")
            {
                mainKeySelect = HKEY_CLASSES_ROOT;
            }
            else if (MainKey.Text == "HKEY_CURRENT_USER")
            {
                mainKeySelect = HKEY_CURRENT_USER;
            }
            else if (MainKey.Text == "HKEY_LOCAL_MACHINE")
            {
                mainKeySelect = HKEY_LOCAL_MACHINE;
            }
            else if (MainKey.Text == "HKEY_USERS")
            {
                mainKeySelect = HKEY_USERS;
            }
        }
        
        private void ValueType_SelectedIndexChanged(object sender, EventArgs e)
        {
            if (ValueType.Text == "String") 
            {
                valueType = REG_SZ;
                ValueLabel.Text = "Value (String)";
                Value.Multiline = false;
            }
            else if (ValueType.Text == "Dword") 
            {
                valueType = REG_DWORD;
                ValueLabel.Text = "Value (Decimal)";
                Value.Multiline = false;
            }
            else if (ValueType.Text == "Binary") 
            {
                valueType = REG_BINARY;
                ValueLabel.Text = "Value (Decimal)";
                Value.Multiline = false;
            }
            else if (ValueType.Text == "Multi String") 
            {
                valueType = REG_MULTI_SZ;
                ValueLabel.Text = "Value (Multi String)";
                Value.Multiline = true;
            }
            Value.Text = "";
        }
        
        private void Form1_closing(object sender, CancelEventArgs e)
        {
            if (allKeysClosed == false)
            {
                RegCloseKey(retNewKey);
                RegCloseKey(retOpenKey);
            }
        }

        private void SaveReg_Click(object sender, EventArgs e)
        {
            if (mainKeySelect == (UIntPtr)0)
            {
                MessageBox.Show("Select main key.", "Error");
            }
            else
            {
                retVal = RegFlushKey(mainKeySelect);
                if (retVal != ERROR_SUCCESS)
                {
                    MessageBox.Show("Error Code = " + retVal, "Error");
                }
                else
                {
                    MessageBox.Show("Main key saved.", "Success");
                }
            }
        }
    }
}